/* emul.c */
/* Only used in TextWin */

#include "config.h"
#include "global.h"
#include "session.h"
#include "cmdparse.h"
#include "emul.h"
#include "pc.h"
#include "vt102.h"

# define UNUSED_ARG(aaa) (void)(aaa)

void   defaultEmulRecvStr __ARGS((struct screen *, char *, int));
void   defaultEmulRecvChar __ARGS((struct screen *, int, int));
int    defaultEmulInit __ARGS((struct screen *));
void   defaultEmulTerm __ARGS((struct screen *));
char * defaultEmulKey __ARGS((struct screen *, int));
void   defaultPrimRecvChar __ARGS((struct screen *, int, int));




void emulGotoXY(scrn, x, y)
struct screen * scrn;
int x;
int y;
{
 scrn->col = x;
 scrn->row = y;

# ifdef TEXTWIN

 TextwinGotoXY(scrn->hWnd, x, y);

# else

	_DH=y;
	_DL=x;
	_BH=0;
	_AH=2;
	geninterrupt(0x10);

# endif
}


void	emulMoveLines(scrn,top,bottom,dest)
struct screen * scrn;
int top;
int bottom;
int dest;
{
	int far *destptr=scrn->videoRAM+dest*scrn->width;
	int far *srcptr=scrn->videoRAM+top*scrn->width;
	int len=(bottom-top+1)*scrn->width*2;
	memmove(destptr,srcptr,len);

# ifdef TEXTWIN
 AddDirtyYRange(scrn->hWnd, top , bottom-top+1);
 AddDirtyYRange(scrn->hWnd, dest, bottom-top+1);
# endif
}

void	emulMoveChars(scrn,line,left,right,dest)
struct screen *scrn;
int line;
int left;
int right;
int dest;
{
	int far *ptr1=scrn->videoRAM+line*scrn->width;
	int far *destptr=ptr1+dest;
	int far *srcptr=ptr1+left;
	int len=(right-left+1)*2;
	memmove(destptr,srcptr,len);

# ifdef TEXTWIN
 AddDirtyY(scrn->hWnd, line);
# endif
}

void emulFillRect(scrn,xLo,yLo,xHi,yHi,filler)
struct screen * scrn;
int xLo;
int yLo;
int xHi;
int yHi;
int filler;
{
 int y;
 int x;
 int far * p;

 for ( y  = yLo;
       y <= yHi;
       y++ )
     {
      p = scrn->videoRAM + scrn->width * y + xLo;

      for ( x  = xLo;
            x <= xHi;
            x++ )
          *p++ = filler;
     }

# ifdef TEXTWIN 
 AddDirtyRect(scrn->hWnd, xLo, yLo, xHi, yHi);
# else
# endif
}

int emulWhereX(scrn)
struct screen * scrn;
{
 return scrn->col;
}

int emulWhereY(scrn)
struct screen * scrn;
{
 return scrn->row;
}

static void ttyScrollUp(struct screen * scrn)
{
 emulMoveLines(scrn, 1, scrn->height-1, 0);
 emulFillRect( scrn, 
               0, 
               scrn->height-1, 
               scrn->width-1, 
               scrn->height-1,

# ifdef TEXTWIN
               ((int)GetWindowAttribute(scrn->hWnd) << 8) | ' '
# else
               0x0720
# endif
             );
}

void defaultPrimRecvChar(scrn, ch, echo)
struct screen * scrn;
int ch;
int echo;
{
 int       cell;
 int far * pCell;

 UNUSED_ARG(echo);

 switch (ch)
        {
         case '\r':
           scrn->col = 0;
           break;

         case '\n':
           scrn->row++;

           if (scrn->row >= scrn->height)
              {
               ttyScrollUp(scrn);
               scrn->row = scrn->height - 1;
              }
           break;

         case '\b':
           scrn->col--;
           if (scrn->col < 0)
              scrn->col = 0;
           break;

         case 7: /* BEL */
           BEGIN_ASM
             mov ah,0eh
             mov al,7
             mov bx,0
             int 10h
           END_ASM
           break;


         case '\t':
           ch = ' ';

           /* DROP THROUGH */

         default:
           pCell = scrn->videoRAM + scrn->width*scrn->row + scrn->col;

           cell = *pCell;
           cell = (cell & 0xFF00) | (ch & 0x00FF);
           *pCell = cell;

# ifdef TEXTWIN
           AddDirtyXY(scrn->hWnd, scrn->col, scrn->row);
# endif

           scrn->col++;
           if (scrn->col >= scrn->width)
              {
               scrn->col = 0;
               scrn->row++;

               if (scrn->row >= scrn->height)
                  {
                   ttyScrollUp(scrn);
                   scrn->row = scrn->height - 1;
                  }
              }

           break;
        }
}


void defaultEmulRecvChar(scrn, ch, echo)
struct screen * scrn;
int ch;
int echo;
{
 defaultPrimRecvChar(scrn, ch, echo);
 emulGotoXY(scrn, scrn->col, scrn->row);

# ifdef TEXTWIN
 UpdateDirtyArea(scrn->hWnd);
# endif
}


void defaultEmulRecvStr(scrn, str, echo)
struct screen * scrn;
char * str;
int echo;
{
 int ch;

 while ((ch = *str++) != 0)
       (*scrn->emul->emulRecvChar)(scrn, ch, echo);

 emulGotoXY(scrn, scrn->col, scrn->row);

# ifdef TEXTWIN
 UpdateDirtyArea(scrn->hWnd);
# endif
}

int defaultEmulInit(scrn)
struct screen * scrn;
{
#ifndef TEXTWIN
 int cols,
     mode;
#endif

 scrn->row = 0;
 scrn->col = 0;

# ifdef TEXTWIN

 scrn->width    = GetTerminalWidth(scrn->hWnd);
 scrn->height   = GetTerminalHeight(scrn->hWnd);
 scrn->videoRAM = (int far *)GetTerminalBuffer(scrn->hWnd);

# else

 scrn->width  = 25;
 scrn->height = 80;

	_AH = 0x0f;
	geninterrupt(0x10);
	cols = _AH;
 mode = _AL;

 scrn->width = cols;

	if (mode == 7)
		  scrn->videoRAM = MK_FP(0xb000, 0);
	else
	   {
		   scrn->videoRAM = MK_FP(0xb800, 0);

		   scrn->height = *(char far *)MK_FP(0x40, 0x84);
		   if (    (scrn->height < 25) 
          || (scrn->height > 60) )
		     	scrn->height = 25;
    }

# endif

 emulFillRect( scrn, 
               0, 
               0, 
               scrn->width-1, 
               scrn->height-1, 
# ifdef TEXTWIN
               ((int)GetWindowAttribute(scrn->hWnd) << 8) | ' '
# else
               0x0720
# endif
             );

 return 1;
}

void defaultEmulTerm(scrn)
struct screen * scrn;
{
 UNUSED_ARG(scrn);
}


char * defaultEmulKey(scrn, key)
struct screen * scrn;
int key;
{
 static char buf[2];

 if ((char)key)
    {
     buf[0] = (char)key;
     buf[1] = '\0';
     
     return buf;
    }
 else
    return NULL;
}


struct emul far Emulations[] =
{
 { "TTY",
   "Dumb TTY",

   defaultEmulInit,
   defaultEmulTerm,
   defaultEmulRecvChar,
   defaultEmulRecvStr,
   defaultEmulKey
 },

# ifdef VT102

 { "VT102",
   "DEC VT102 (24 rows)",

   init_vt102,
   term_vt102,
   vt102_putc,
   vt102_puts,
   translate_vt102
 },

 { "BigVT102",
   "DEC VT102 (Full screen)",

   init_big_vt102,
   defaultEmulTerm,
   vt102_putc,
   vt102_puts,
   translate_vt102
 },

# endif

 { NULL
 }
};



struct emul * far DefSessionEmulation[] = /* In session-type order */
{
 &Emulations[EMUL_TTY],   /* Free      */

# ifdef VT102
 &Emulations[EMUL_VT102], /* TELNET 	  */
# else
 &Emulations[EMUL_TTY],   /* TELNET 	  */
# endif                  

 &Emulations[EMUL_TTY],   /* FTP    	  */
 &Emulations[EMUL_TTY],   /* AX25TNC	  */
 &Emulations[EMUL_TTY],   /* FINGER	   */
 &Emulations[EMUL_TTY],   /* PING	     */
 &Emulations[EMUL_TTY],   /* NRSESSION */
 &Emulations[EMUL_TTY],   /* COMMAND	  */
 &Emulations[EMUL_TTY],   /* MORE      */
 &Emulations[EMUL_TTY],   /* HOP	      */
 &Emulations[EMUL_TTY],   /* TIP	      */
 &Emulations[EMUL_TTY],   /* PPPPASS	  */
 &Emulations[EMUL_TTY],   /* DIAL	     */
 &Emulations[EMUL_TTY],   /* DQUERY	   */
 &Emulations[EMUL_TTY],   /* DCLIST	   */
 &Emulations[EMUL_TTY],   /* RLOGIN	   */

};


struct emul * getSessionDefEmulation(sp)
struct session * sp;
{
 if ((unsigned int)sp->type >= sizeof(DefSessionEmulation) / sizeof(struct emul *))
    return &Emulations[EMUL_TTY];
 else
    return DefSessionEmulation[sp->type];
}


struct emul * findEmulation(name)
char * name;
{
 struct emul * p = Emulations;
 int len = strlen(name);

 while (p->name != NULL)
       {
        if (strnicmp(p->name, name, len) == 0)
           return p;

        p++;
       }

 return NULL;
}



int
doemulnoop(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(argc);
 UNUSED_ARG(argv);
 UNUSED_ARG(p);

 return 0;
}


int
doemuldefault(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 unsigned int type = (unsigned int)atoi(argv[1]);
 struct emul * emul;

 UNUSED_ARG(p);

 if (     type > 0
      &&  type < sizeof(DefSessionEmulation) / sizeof(struct emul *))
    if (argc <= 2)
       tprintf( "def emul for sess type %d (%s): %s\n",
                type,
                Sestypes[type],
                DefSessionEmulation[type]->name );
    else
       {
        emul = findEmulation(argv[2]);
    
        if (emul != NULL)
           DefSessionEmulation[type] = emul;
       }

 return 0;
}


int
doemullist(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 struct emul * emul = Emulations;

 UNUSED_ARG(argc);
 UNUSED_ARG(argv);
 UNUSED_ARG(p);

 tprintf("Terminal emuls:\n");

 while (emul->name != NULL)
       {
        tprintf(" %s: %s\n", emul->name, emul->desc);

        emul++;
       }

 return 0;
}


int
doemulsession(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 struct session * sp = sessptr(argv[1]);
 struct emul * emul;

 UNUSED_ARG(p);

 if (sp != NULLSESSION)
    {
     if (argc <= 2)
        tprintf( "Sess %d emul: %s\n", 
                 atoi(argv[1]), 
                 sp->screen->emul->name );
     else
        {
         emul = findEmulation(argv[2]);

         if (emul != NULL)
            {
             (*sp->screen->emul->emulTerm)(sp->screen);

             sp->screen->emul = emul;

             (*sp->screen->emul->emulInit)(sp->screen);
            }
        }
    }

 return 0;
}


struct cmds far Emulationcmds[] = {
    { "",         doemulnoop,    0, 0, NULLCHAR },
    { "default",  doemuldefault, 0, 2, "default <sess-type> <emul>" },
    { "list",     doemullist,    0, 0, NULLCHAR },
    { "session",  doemulsession, 0, 1, "session <sess> [<emul>]"},
    { NULLCHAR,   NULLFP,	 0, 0, NULLCHAR }
};



int
doemulation(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 return subcmd(Emulationcmds,argc,argv,p);
}




