/* textwine.c */

/*
   Routines to help interface with terminal emulators
*/

#include "config.h"
#include "global.h"
#include "textwin.h"
#include "textwinp.h"



int     GetTerminalWidth(HWINDOW pWindow)
{
 return WINLOGICALWIDTH(pWindow);
}


int     GetTerminalHeight(HWINDOW pWindow)
{
 return WINLOGICALHEIGHT(pWindow);
}


unsigned char* GetTerminalBuffer(HWINDOW pWindow)
{
 return WINLOGICALBUFFER(pWindow);
}


void    TextwinGotoXY(HWINDOW pWindow, int x, int y)
{
 int dx = 0,
     dy = 0;

 if (IsWindow(pWindow))
    {
     x = max(0, min(x, WINLOGICALWIDTH( pWindow)-1));
     y = max(0, min(y, WINLOGICALHEIGHT(pWindow)-1));
     
     WINCURX(pWindow) = x;
     WINCURY(pWindow) = y;
     
     if (autoScrollToCursor)
        {
         if (autoScrollToCursor & 0x01)
            {
             if (y < WINLOGICALY(pWindow))
                dy = WINLOGICALY(pWindow) - y;
             else if (y >= WINLOGICALY(pWindow) + WINCLIENTHEIGHT(pWindow))
                dy = WINLOGICALY(pWindow) + WINCLIENTHEIGHT(pWindow) - y - 1;
            }

         if (autoScrollToCursor & 0x02)
            {
             if (x < WINLOGICALX(pWindow))
                dx = WINLOGICALX(pWindow) - x;
             else if (x >= WINLOGICALX(pWindow) + WINCLIENTWIDTH(pWindow))
                dx = WINLOGICALX(pWindow) + WINCLIENTWIDTH(pWindow) - x - 1;
            }

         if (dx || dy)
            {
             WINLOGICALX(pWindow) -= dx;
             WINLOGICALY(pWindow) -= dy;

             UpdateWindow(pWindow);
            }
        }
     
     if (pWindow == focuswindow)
        UpdateCursorDisplay(pWindow);
    }
}


int     GetCursorX(HWINDOW pWindow)
{
 return WINCURX(pWindow);
}

int     GetCursorY(HWINDOW pWindow)
{
 return WINCURY(pWindow);
}




void ResetDirtyArea(HWINDOW pWindow)
{
 SetWinDirtyLoX(pWindow,  32767);
 SetWinDirtyHiX(pWindow,  -32767);
 SetWinDirtyLoY(pWindow,   32767);
 SetWinDirtyHiY(pWindow,  -32767);
}

void UpdateDirtyArea(HWINDOW pWindow)
{
 if (GetWinDirtyLoX(pWindow) <= GetWinDirtyHiX(pWindow))
    if (GetWinDirtyLoY(pWindow) <= GetWinDirtyHiY(pWindow))
       {
        UpdateLogicalRect( pWindow, 
                           GetWinDirtyLoX(pWindow),
                           GetWinDirtyHiX(pWindow),
                           GetWinDirtyLoY(pWindow),
                           GetWinDirtyHiY(pWindow) );

        ResetDirtyArea(pWindow);
       }
}


void AddDirtyXY(pWindow, x, y)
HWINDOW pWindow;
int x;
int y;
{
 SetWinDirtyLoX(pWindow, min(GetWinDirtyLoX(pWindow), x));
 SetWinDirtyHiX(pWindow, max(GetWinDirtyHiX(pWindow), x));

 SetWinDirtyLoY(pWindow, min(GetWinDirtyLoY(pWindow), y));
 SetWinDirtyHiY(pWindow, max(GetWinDirtyHiY(pWindow), y));
}


void AddDirtyY(pWindow, y)
HWINDOW pWindow;
int y;
{
 SetWinDirtyLoX(pWindow, 0);
 SetWinDirtyHiX(pWindow, WINLOGICALWIDTH(pWindow) - 1);

 SetWinDirtyLoY(pWindow, min(GetWinDirtyLoY(pWindow), y));
 SetWinDirtyHiY(pWindow, max(GetWinDirtyHiY(pWindow), y));
}

void AddDirtyCurXY(pWindow)
HWINDOW pWindow;
{
 AddDirtyXY( pWindow,
             WINCURX(pWindow),
             WINCURY(pWindow) );
}

void AddDirtyCurY(pWindow)
HWINDOW pWindow;
{
 AddDirtyY( pWindow, WINCURY(pWindow));
}

void AddDirtyYRange(pWindow, y, dy)
HWINDOW pWindow;
int y;
int dy;
{
 AddDirtyXY(pWindow, 0, y);
 AddDirtyXY(pWindow, WINLOGICALWIDTH(pWindow)-1, y+dy-1);
}

void AddDirtyRect(pWindow, x0, y0, x1, y1)
HWINDOW pWindow;
int x0;
int y0;
int x1;
int y1;
{ 
 AddDirtyXY(pWindow, x0, y0);
 AddDirtyXY(pWindow, x1, y1);
}


void PutDirtyCharAt(HWINDOW pWindow, int x, int y, char ch)
{
 unsigned char* p;

 if (y >= 0  &&  y < WINLOGICALHEIGHT(pWindow))
    if (x >= 0  &&  x < WINLOGICALWIDTH(pWindow))
       {
        p = &WINCHARXY(pWindow, x, y);

        *p = ch;

        AddDirtyXY(pWindow, x, y);
       }
}

void PutDirtyStrAt(HWINDOW pWindow, int x, int y, PSZ str)
{
 unsigned char* p;
 char ch;

 if (y >= 0  &&  y < WINLOGICALHEIGHT(pWindow))
    {
     p = &WINCHARXY(pWindow, x, y);

     while ((ch = *str++) != '\0')
           {
            if (     x >= 0  
                 &&  x <  WINLOGICALWIDTH(pWindow) )
               {
                *p = ch;
               
                AddDirtyXY(pWindow, x, y);
               }
               
            p += 2;
            x++;
           }
    }
}



/* eof: textwine.c */
