/* textwinm.c */

/*
   Textwin mouse & message stuff...
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <alloc.h>
#include "config.h"
#include "global.h"
#include "files.h"
#include "proc.h"
#include "session.h"
#include "textwin.h"
#include "textwinp.h"


# ifdef TEXTWIN



typedef unsigned long DWORD;
typedef DWORD dword;
typedef WORD word;
typedef BYTE byte;

/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                M o u s e                                 */
/*                                                                          */
/*--------------------------------------------------------------------------*/

# define ShowHardMouse() { BEGIN_ASM mov ax,1 END_ASM BEGIN_ASM int 33h END_ASM }
# define HideHardMouse() { BEGIN_ASM mov ax,2 END_ASM BEGIN_ASM int 33h END_ASM }
 

static int          oldMouseX = 0;
static int          oldMouseY = 0;
static int          mouseIsVisible = 0;



# define WIBBLE(x,y) \
{if ((x)>=0 && (x)<screenWidth && (y)>=0 && (y)<screenHeight)\
{*(ptr + 2 * ((y)*screenWidth + (x)) + 1) ^= (BYTE)param[TW_PARAM_WIBBLE_FACTOR];}}


void MoveMouse(int mouseX, int mouseY)
{
 PBYTE ptr = screenBuffer;

 if (mouseIsVisible)
    if (mouseKind == 1)
       {
        WIBBLE(oldMouseX, oldMouseY);
    
        WIBBLE(mouseX, mouseY);
    
        oldMouseX = mouseX;
        oldMouseY = mouseY;
       }
}


void ShowMouse(void)
{
 PBYTE ptr = screenBuffer;

 mouseIsVisible++;

 if (mouseIsVisible)
    if (mouseKind == 1)
       WIBBLE(oldMouseX, oldMouseY)
    else
       ShowHardMouse();
}


void HideMouse(void)
{
 PBYTE ptr = screenBuffer;

 if (mouseIsVisible)
    if (mouseKind == 1)
       WIBBLE(oldMouseX, oldMouseY)
    else
       HideHardMouse();

 mouseIsVisible--;
}


typedef enum
{
 MSG_NULL         = 0,
 MSG_LBUTTONDOWN,
 MSG_LBUTTONUP,
 MSG_RBUTTONDOWN,
 MSG_RBUTTONUP,
 MSG_MOUSEMOVE,
 MSG_CHAR,

 MSG_SIZEWINDOW,  // AMcC : 27Dec92
 MSG_MOVEWINDOW,  // AMcC : 27Dec92
 MSG_NEXTWINDOW   // AMcC : 27Dec92

} MessageType;


typedef enum
{
 WSUBMSG_NULL = 0,
 WSUBMSG_LEFT,
 WSUBMSG_RIGHT,
 WSUBMSG_UP,
 WSUBMSG_DOWN,
} WindowSubMessageType;

typedef struct tagMessage FAR * LPMessage;
typedef volatile struct tagMessage FAR * LPVolatileMessage;
typedef struct tagMessage
{
 MessageType message;    /* One of the above Messages */
 long time;     /* Time (in ticks) the message was queued */

 union
 {
   struct
   {
    WORD xpos;
    WORD ypos;
    WORD mouseState;
    WORD shiftState;
   } mouse;

   struct 
   {
    char c;
    BYTE scanCode;
    WORD shiftState;
   } key;

   struct 
   {
    WORD  id;
   } timer;

   struct
   {
    int submessage;
   } window;

 } info;

} Message;


# define MAX_MESSAGE_QUEUE (64)


typedef struct tagMessageQueue FAR * LPMessageQueue;
typedef struct tagMessageQueue
{
 WORD nextIn;
 WORD nextOut;

 Message msg[MAX_MESSAGE_QUEUE];
} MessageQueue;


volatile MessageQueue messageQueue = {0,0};



# define IsEmptyMessageQueue(mq) ((mq).nextIn == (mq).nextOut)

# define IsFullMessageQueue(mq) (  (mq).nextIn == (mq).nextOut - 1 \
                                 ||  (mq).nextIn  == MAX_MESSAGE_QUEUE - 1 \
                                   &&(   (mq).nextOut == 0 \
                                      || (mq).nextOut >= MAX_MESSAGE_QUEUE \
                                     ))


void AddMessage(LPMessage pmsg)
{
 LPVolatileMessage p;
 WORD prev;

 if (pmsg->message == MSG_MOUSEMOVE)
    if (! IsEmptyMessageQueue(messageQueue))
       {
        if (messageQueue.nextIn > 0)
           prev = messageQueue.nextIn - 1;
        else
           prev = MAX_MESSAGE_QUEUE - 1;
    
        if (messageQueue.msg[prev].message == MSG_MOUSEMOVE)
           messageQueue.nextIn = prev;
       }

 if (! IsFullMessageQueue(messageQueue))
    {
     /*
        If the last message in the queue is a mouse move message
        and this new message is also a mouse move message, we
        overwrite the previous one. This avoids the very likely
        situation of the queue becoming full of mouse move messages
        which the app has no chance to process.
     */
     
     p = &messageQueue.msg[messageQueue.nextIn++];
     messageQueue.nextIn %= MAX_MESSAGE_QUEUE;

     *p = *pmsg;

     p->time = 0;
    }

 psignal(&messageQueue, 1);
}


int GetMessage(LPMessage m)
{
 int bGotMsg = 0;

 if (! IsEmptyMessageQueue(messageQueue))
    {
     /*
        Disable interrupts while removing message from queue.
     */

     _disable();

     *m = messageQueue.msg[messageQueue.nextOut++];
     messageQueue.nextOut %= MAX_MESSAGE_QUEUE;

     _enable();

     bGotMsg = TRUE;
    }

 return bGotMsg;
}


void FAR mouseRoutine(void)
{
 static int dsVar;
 Message msg;
 int r_ax;

 BEGIN_ASM
       mov r_ax,ax

       mov word ptr msg.info.mouse.xpos, cx
       mov word ptr msg.info.mouse.ypos, dx
       mov word ptr msg.info.mouse.mouseState, bx

       push ds

       mov ax,seg dsVar
       mov ds,ax

       mov ah,2
       int 16h
       mov word ptr msg.info.mouse.shiftState, ax

 END_ASM

 msg.message = MSG_NULL;

 if (r_ax & 0x0001)
    msg.message = MSG_MOUSEMOVE;

 if (r_ax & 0x0002)
    msg.message = MSG_LBUTTONDOWN;
 else if (r_ax & 0x0004)
    msg.message = MSG_LBUTTONUP;
 else if (r_ax & 0x0008)
    msg.message = MSG_RBUTTONDOWN;
 else if (r_ax & 0x0010)
    msg.message = MSG_RBUTTONUP;
 
 if (msg.message != MSG_NULL)
    {
     msg.info.mouse.xpos >>= 3;
     msg.info.mouse.ypos /= screenCharHeight;
     
     msg.info.mouse.xpos = min(msg.info.mouse.xpos, screenWidth -1);
     msg.info.mouse.ypos = min(msg.info.mouse.ypos, screenHeight-1);
     
     AddMessage(&msg);
    }

 BEGIN_ASM
       mov cx,0

       pop ds
 END_ASM
}

static HWINDOW trackWindow = NULL;


static int CalcOrgFromScrollThumbPos( int sBeg,
                                      int sEnd,
                                      int sMin,
                                      int sMax,
                                      int sPos )
{
 int  sVal;
 long cells;

 sPos = max(sBeg, min(sPos, sEnd));

 if (sPos <= sBeg)
    sVal = sMin;
 else if (sPos >= sEnd)
    sVal = sMax;
 else  
    {
     cells = sEnd - sBeg /* + 1 */;

     sVal = sMin + (int)(((long)(sMax - sMin + 1) * (sPos - sBeg)) / cells);

     sVal = max(sMin, min(sVal, sMax));
    }

 return sVal;
}


void ScrollGadgetHit(HWINDOW pWindow, HitType ht, int mouX, int mouY)
{
 int newX = WINLOGICALX(pWindow),
     newY = WINLOGICALY(pWindow);
 int cliW = WINCLIENTWIDTH(pWindow);
 int cliH = WINCLIENTHEIGHT(pWindow);

 switch (ht)
        {
         case HT_SCROLL_LEFT:      newX--; break;
         case HT_SCROLL_RIGHT:     newX++; break;
         case HT_SCROLL_UP:        newY--; break;
         case HT_SCROLL_DOWN:      newY++; break;

         case HT_SCROLL_PAGELEFT:  newX -= max(1, 3 * cliW / 4); break;
         case HT_SCROLL_PAGERIGHT: newX += max(1, 3 * cliW / 4); break;
         case HT_SCROLL_PAGEUP:    newY -= max(1, 3 * cliH / 4); break;
         case HT_SCROLL_PAGEDOWN:  newY += max(1, 3 * cliH / 4); break;

         case HT_SCROLL_HORZTHUMB: 
           newX = CalcOrgFromScrollThumbPos( 2,
                                             WINWIDTH(pWindow)-3,
                                             0,
                                             WINLOGICALWIDTH(pWindow)
                                              - cliW,
                                             mouX - WINLEFT(pWindow) );
           break;

         case HT_SCROLL_VERTTHUMB: 
           newY = CalcOrgFromScrollThumbPos( 2,
                                             WINHEIGHT(pWindow)-3,
                                             0,
                                             WINLOGICALHEIGHT(pWindow)
                                              - cliH,
                                             mouY - WINTOP(pWindow) );
           break;
        }

 newX = max(0, min(newX, WINLOGICALWIDTH( pWindow) - cliW));
 newY = max(0, min(newY, WINLOGICALHEIGHT(pWindow) - cliH));

 if (    newX != WINLOGICALX(pWindow)
      || newY != WINLOGICALY(pWindow) )
    {
     WINLOGICALX(pWindow) = newX;
     WINLOGICALY(pWindow) = newY;

     UpdateCursorDisplay(pWindow);
     UpdateWindow(pWindow);
     FrameWindow(pWindow);
    }
}


static void SetWindowFocusAndCurrentSession(HWINDOW w)
{
 struct session * sp;

 SetWindowFocus(w);

 sp = (struct session *)GetWindowUserData(w);
 if (sp != NULLSESSION  &&  Current != sp)
    {
     if (sp != Command) /* AMcC : 9Jul92 : remember last session */
        Lastcurr = sp;

	    swapscreen(Current,sp);
	    Current = sp;
	    psignal(sp,0);
     
	    Current->row = MOREROWS;
	    psignal(&Current->row,1);
    }

}

void TextwinMessageDispatcher(Message * msg)
{
 static int     xoff, 
                yoff;
 static HitType ht;

 HWINDOW w;
 int x,
     y;
 SHORT newX,
       newY;


 TextwinInit();

 if (bTextwinInitialized)
    {
     x = msg->info.mouse.xpos,
     y = msg->info.mouse.ypos;

     switch (msg->message)
            {
             case MSG_LBUTTONDOWN:
               w = WindowFromPoint(x, y);

               if (w != NULL)
                  if (w != GetDesktopWindow())
                     {
                      SetWindowFocusAndCurrentSession(w);

                      ht = WindowHitType(w, x, y);

                      xoff = x - WINLEFT(w);
                      yoff = y - WINTOP(w);

                      if (    ht == HT_CAPTION
                           || ht == HT_SIZE    )
                         trackWindow = w;
                      else if (    ht == HT_MAXIMIZE
                                || ht == HT_MINIMIZE
                                || ht == HT_RESTORE  )
                         ZoomWindow(w, ht);
                      else if (ht >= HT_SCROLL_LEFT  &&  ht <= HT_SCROLL_PAGEDOWN)
                         ScrollGadgetHit(w, ht, x, y);      
                      else if (    ht == HT_SCROLL_VERTTHUMB
                                || ht == HT_SCROLL_HORZTHUMB )
                         trackWindow = w;


                      pwait(NULL); // Yield
                     }
               break;

             case MSG_MOUSEMOVE:
               MoveMouse(x, y);

               if (trackWindow != NULL)
                  {
                   if (ht == HT_CAPTION)
                      {
                       newX = x - xoff;
                       newY = y - yoff;
                       
                       if (    newX != WINXPOS(trackWindow)
                            || newY != WINYPOS(trackWindow) )
                          WindowSetPos(trackWindow, newX, newY);
                      }
                   else if (ht == HT_SIZE)
                      {
                       newX = max(5, x - WINLEFT(trackWindow) + 1);
                       newY = max(2, y - WINTOP(trackWindow) + 1);

                       if (    newX != WINWIDTH(trackWindow)
                            || newY != WINHEIGHT(trackWindow) )
                          ResizeWindow(trackWindow, newX, newY);
                      }
                   else if (    ht == HT_SCROLL_VERTTHUMB
                             || ht == HT_SCROLL_HORZTHUMB )
                      ScrollGadgetHit(trackWindow, ht, x, y);

                   pwait(NULL);
                  }

               break;

             case MSG_LBUTTONUP:
               if (trackWindow != NULL)
                  trackWindow = NULL;
               break;


             // AMcC : 27Dec92

             case MSG_MOVEWINDOW:
               w = GetWindowFocus();

               if (    w != NULL
                    && w != GetDesktopWindow() )
                  {
                   newX = WINXPOS(w);
                   newY = WINYPOS(w);

                   switch (msg->info.window.submessage)
                          {
                           case WSUBMSG_LEFT : newX--; break;
                           case WSUBMSG_RIGHT: newX++; break;
                           case WSUBMSG_UP   : newY--; break;
                           case WSUBMSG_DOWN : newY++; break;
                          }

                   WindowSetPos(w, newX, newY);
                  }

               break;

             case MSG_SIZEWINDOW:
               w = GetWindowFocus();

               if (    w != NULL
                    && w != GetDesktopWindow() )
                  {
                   newX = WINWIDTH(w);
                   newY = WINHEIGHT(w);

                   switch (msg->info.window.submessage)
                          {
                           case WSUBMSG_LEFT : newX--; break;
                           case WSUBMSG_RIGHT: newX++; break;
                           case WSUBMSG_UP   : newY--; break;
                           case WSUBMSG_DOWN : newY++; break;
                          }

                   newX = max(5, newX);
                   newY = max(2, newY);

                   ResizeWindow(w, newX, newY);
                  }
               break;

             case MSG_NEXTWINDOW:
               w = GetWindowBottom();

               if (IsWindow(w))
                  if (w != GetWindowTop())
                     SetWindowFocusAndCurrentSession(w);
               break;
            }
    }
}


/* Mouse input process */
void
mouseProcess(i,v1,v2)
int i;
void *v1;
void *v2;
{
 Message msg;
     
 UNUSED_ARG(i);
 UNUSED_ARG(v1);
 UNUSED_ARG(v2);

 /* Mouse process loop */
 for (;;)
     {
      while (IsEmptyMessageQueue(messageQueue))
            pwait(&messageQueue);
     
      while (GetMessage(&msg))
            TextwinMessageDispatcher(&msg);
     }
}


// AMcC : 27Dec92

typedef struct tagKeyBinding
{
 int                  key1,        // 1st char of key seq: (0 for ext key)
                      key2;        // 2nd char of key seq: (ignored if key1
                                   //                       is non-zero)

 MessageType          message;     // message/submessage queued when the
 WindowSubMessageType submessage;  //  specified key sequence is recognised
} KeyBinding;


static KeyBinding sysKey[] =
{
 { 0, 0x9b, MSG_MOVEWINDOW, WSUBMSG_LEFT  },
 { 0, 0x9d, MSG_MOVEWINDOW, WSUBMSG_RIGHT },
 { 0, 0x98, MSG_MOVEWINDOW, WSUBMSG_UP    },
 { 0, 0xa0, MSG_MOVEWINDOW, WSUBMSG_DOWN  },

 { 0, 0x73, MSG_SIZEWINDOW, WSUBMSG_LEFT  },
 { 0, 0x74, MSG_SIZEWINDOW, WSUBMSG_RIGHT },
 { 0, 0x8d, MSG_SIZEWINDOW, WSUBMSG_UP    },
 { 0, 0x91, MSG_SIZEWINDOW, WSUBMSG_DOWN  },

 { 0, 0x6d, MSG_NEXTWINDOW, WSUBMSG_NULL  },

 {0,0,0,0} // Terminator, code assumes this is here
};


int dotwkey(argc,argv,p)
int argc;
char **argv;
void * p;
{
 int which = atoi(argv[1]);
 int qty   = (sizeof(sysKey) / sizeof(sysKey[0])) - 1;
 int key1  = 0,
     key2  = 0;

 UNUSED_ARG(p);

 if (which < 0  ||  which >= qty)
    tprintf("Bad key index\n");
 else
    if (argc <= 2)
       tprintf( "key %d: %d %d\n", 
                which, 
                sysKey[which].key1, 
                sysKey[which].key2 );
    else
       {
        if (argc >= 4)
           key2 = atoi(argv[3]);

        key1 = atoi(argv[2]);

        sysKey[which].key1 = key1;
        sysKey[which].key2 = key2;
       }

 return 0;
}


int IsTextwinSystemKeystroke(int key)
{
 int bHandled = 0;
 Message msg;
 int ctr;
 int keyLo = (key     ) & 0xff,
     keyHi = (key >> 8) & 0xff;

 ctr = 0;
 while (     ! bHandled
         &&  sysKey[ctr].message  )
       {
        if (    (    keyLo == 0x00
                  && sysKey[ctr].key1 == keyLo
                  && sysKey[ctr].key2 == keyHi )
             ||    
                (    keyLo != 0x00
                  && sysKey[ctr].key1 == keyLo )
           )
           {
            msg.message                = sysKey[ctr].message;
            msg.info.window.submessage = sysKey[ctr].submessage;

            bHandled = 1;
           }

        ctr++;
       }

 if (bHandled)
    AddMessage(&msg);

 return bHandled;
}


# endif /* TEXTWIN */


/* eof: textwinm.c */

