/* textwinp.h */

/*
   Stuff in here is private to the textwin*.c modules
*/

# ifndef TEXTWINP_H 
# define TEXTWINP_H 


/*
** If you #define AMC_MEMORY_PARANOIA, extensive memory testing will
** occur whenever a window is accessed. This allows rapid detection
** of trampling, hopefully before things get _too_ serious.
*/

// # define AMC_MEMORY_PARANOIA 1


typedef enum
{
 TW_PARAM_NULL,          // unused

 TW_PARAM_DESKTOP_CHAR,  // the desktop erase character
 TW_PARAM_WIBBLE_FACTOR, // the software mouse wibble-factor

 TW_PARAM_LAST           // unused
} TwParamIndex;


typedef enum
{
 HT_NULL = 0,
 HT_CAPTION,
 HT_CLIENT,
 HT_SIZE,
 HT_CLOSE,
 HT_MAXIMIZE,
 HT_MINIMIZE,
 HT_RESTORE,

 HT_SCROLL_VERTTHUMB,
 HT_SCROLL_HORZTHUMB,

 HT_SCROLL_LEFT,
 HT_SCROLL_RIGHT,
 HT_SCROLL_UP,
 HT_SCROLL_DOWN,
 HT_SCROLL_PAGELEFT,
 HT_SCROLL_PAGERIGHT,
 HT_SCROLL_PAGEUP,
 HT_SCROLL_PAGEDOWN
} HitType;


#define WINDOWMAGIC 0xBEEF

# ifndef abs
#   define abs(x)   (((x)<0) ? -(x) : (x))
# endif

# ifndef min
#   define min(x,y) ((x)<(y) ? (x) : (y))
# endif

# ifndef max
#   define max(x,y) ((x)>(y) ? (x) : (y))
# endif


typedef struct 
{
 SHORT xLeft;
 SHORT yTop;
 SHORT width;
 SHORT height;
} SCREENAREASTRUCT;
typedef SCREENAREASTRUCT FAR * PSCREENAREA;


// Visibility buffer pointer
// This points to an array of bytes which describes the
// visibility of each cell of the window.
// Each bit of the visibility buffer is associated with
// one cell of the window. If the bit is set, the cell
// is visible.
// Bit 7 (the MSB) of the zeroth byte in the visBuffer is
// associated with the top-left cell of the window (cell 0).
// Bit6=cell1, bit5=cell2, etc, bit0=cell7, byte1bit7=cell8, etc etc.
// The first cell of each line is associated with a new byte
// in the visBuffer. Thus there may be between 0 and 7 wasted
// bits in each line of the visBuffer.

typedef PBYTE PVISBUFFER;

typedef BYTE               VISFLAGSTYPE;
typedef VISFLAGSTYPE FAR * PVISFLAGS;

#define VFL_NORMAL           0x00
#define VFL_WHOLELINEVISIBLE 0x01
#define VFL_WHOLELINEHIDDEN  0x02

typedef struct {
          PVISBUFFER visbuffer;     // The window visibility buffer
          PVISFLAGS  visflags;      // Visibility line flags
          USHORT     viswidth;      // Bytes per visibility line
          USHORT     visheight;     // Lines of visibility info
          USHORT     visalign;      // Bit where the visbuffer lines begin
          USHORT     visbitswide;   // Number of significant bits per line
        } VISIBILITYSTRUCT;
typedef VISIBILITYSTRUCT FAR * PVISIBILITY;

// Macros to manipulate fields of the visibility structure
// v is a VISIBILITYSTRUCT

#define VISWIDTH(v)                ((v).viswidth)
#define VISBITSWIDE(v)             ((v).visbitswide)
#define VISHEIGHT(v)               ((v).visheight)
#define VISALIGN(v)                ((v).visalign)
#define VISBUFFER(v)               ((v).visbuffer)
#define VISFLAGS(v)                ((v).visflags)
#define VISLINEFLAGS(v,y)          (VISFLAGS(v)[y])
#define VISLINE(v,y)               ((PBYTE)&(VISBUFFER(v)[(y)*VISWIDTH(v)]))
#define VISX(v,x)                  ((x)+VISALIGN(v))
#define VISBYTEOFFSET(v,x)         (VISX(v,x) / 8)
#define VISBYTE(v,x,y)             (VISLINE(v,y)[VISBYTEOFFSET(v,x)])
#define VISWHOLELINEISVISIBLE(v,y) (VISLINEFLAGS(v,y) & VFL_WHOLELINEVISIBLE)
#define VISWHOLELINEISHIDDEN(v,y)  (VISLINEFLAGS(v,y) & VFL_WHOLELINEHIDDEN)

// Utility macros to calculate the visibility buffer width
// and height, given a window's width and height in character cells

#define CALCVISWIDTH(w)   (((w)+15) >> 3)
#define CALCVISBYTES(w,h) (h * CALCVISWIDTH(w))
#define CALCVISFLAGS(w,h) (h * sizeof(VISFLAGSTYPE))

typedef struct {
          BYTE vCh ;          // Simple vertical line (|)
          BYTE hCh ;          // Simple horizontal line (-)
          BYTE tlCh;          // Top left corner (/)
          BYTE blCh;          // Bottom left corner (\)
          BYTE trCh;          // Top right corner (\)
          BYTE brCh;          // Bottom right corner (/)
          BYTE xCh;           // Cross char (+)
          BYTE jlCh;          // Left junction (-|)
          BYTE jrCh;          // Right junction (|-)
          BYTE juCh;          // Up the junction (inverted T)
          BYTE jdCh;          // Down junction (T)

          BYTE resCh;     // restore Char
          BYTE maxCh;     // maximize Char
          BYTE minCh;     // minimize Char
          BYTE lscrCh;    // left scroll
          BYTE rscrCh;    // right scroll
          BYTE uscrCh;    // up scroll
          BYTE dscrCh;    // down scroll
          BYTE hthumbCh;  // horizontal thumb
          BYTE vthumbCh;  // vertical thumb
          BYTE hscrbarCh; // horizontal scroll bar
          BYTE vscrbarCh; // vertical scroll bar

        } FRAMECHARSSTRUCT;

#define WFL_NORMAL       0x0000
#define WFL_NOVERTBORDER 0x0001
#define WFL_NOHORZBORDER 0x0002
#define WFL_NOBORDER     0x0003 /* Mask : NOHORZ | NOVERT */
#define WFL_HIDDEN       0x0004
#define WFL_MAXIMIZED    0x0008
#define WFL_MINIMIZED    0x0010

typedef struct tagWINDOWSTRUCT 
{
 USHORT           magic;         // Must be WINDOWMAGIC to be a window
 SCREENAREASTRUCT windowarea;    // Dimensions and location of window
 SCREENAREASTRUCT clientarea;    // Dimensions and location of client
 USHORT           flags;         // Misc flags
 USHORT           id;            // Who am I ? (Zero is not valid)
 USHORT           disableUpdate; // Disable update count
 USHORT           defaultAttr;   // Attribute to use when none is specified
 BYTE             backgroundCh;  // Background char (normally space)

 // nextWindow and prevWindow describe the Z-Order of the
 // windows. nextWindow is below this window. prevWindow
 // is above this window. If prevWindow is NULL then this
 // window is at the top.
 struct tagWINDOWSTRUCT FAR * nextWindow;    // Next window in list
 struct tagWINDOWSTRUCT FAR * prevWindow;    // Previous window in list

 VISIBILITYSTRUCT vis;
 SHORT            curX;          // Current position in client area
 SHORT            curY;
 FRAMECHARSSTRUCT frame;         // The frame characters

 PSZ              caption;       // Window caption
 void FAR *       userData;      // Arbitrary data

 SCREENAREASTRUCT restorearea;   // Location of window before it
                                 // was maximized

 BYTE             borderAttr;    // Attribute for the border
 PBYTE            borderBuffer;  // Scratch space for the caption, etc

 PBYTE            logicalBuffer;  // A pointer to a block of memory
                                  // that contains the chars and
                                  // attributes of the logical screen

 int logicalX, // Coordinates of the top,left character in the
     logicalY; // window relative to the top,left of the logical
               // screen.  Thus, if the window is showing logical
               // column 2 in window column 0, logicalX will be
               // (positive) 2.

 int logicalWidth,  // Height and width of the logical screen
     logicalHeight;

 int dirtyLoX, // Terminal dirty region. There is no dirty region
     dirtyHiX, // if either dirtyLo? is greater then the dirtyHi?
     dirtyLoY, // value. Coordinates are INCLUSIVE. eg (0,0,0,0)
     dirtyHiY; // specifies a dirty area of 1 character.

} WINDOWSTRUCT;
typedef struct tagWINDOWSTRUCT FAR * PW;


// Macros to produce bitmasks for use with bytes
//
// BYTEBIT() macros produce bit masks where the parameter
// is the offset of the bit from the LSB (normal usage)
#define BYTEBIT(n)                  ((BYTE)(0x01 << (n)))
#define NOTBYTEBIT(n)               ((BYTE)~BYTEBIT(n))


// A few macros to simplify access to window structure elements.
// In all instances, p is either an HWINDOW or a PW (same difference really).

#define WINFRAMECHARS(p)           (((PW)(p))->frame)
#define WINFRAMECHV(p)             (((PW)(p))->frame.vCh)
#define WINFRAMECHH(p)             (((PW)(p))->frame.hCh)
#define WINFRAMECHTL(p)            (((PW)(p))->frame.tlCh)
#define WINFRAMECHBL(p)            (((PW)(p))->frame.blCh)
#define WINFRAMECHTR(p)            (((PW)(p))->frame.trCh)
#define WINFRAMECHBR(p)            (((PW)(p))->frame.brCh)
#define WINFRAMECHX(p)             (((PW)(p))->frame.xCh)
#define WINFRAMECHJL(p)            (((PW)(p))->frame.jlCh)
#define WINFRAMECHJR(p)            (((PW)(p))->frame.jrCh)
#define WINFRAMECHJU(p)            (((PW)(p))->frame.juCh)
#define WINFRAMECHJD(p)            (((PW)(p))->frame.jdCh)

#define WINNEXTWINDOW(p)           (((PW)(p))->nextWindow)
#define WINPREVWINDOW(p)           (((PW)(p))->prevWindow)
#define WINISTOPWINDOW(p)          (! WINPREVWINDOW(p))
#define WINISBOTWINDOW(p)          (! WINNEXTWINDOW(p))
#define WINDISABLEUPDATE(p)        (((PW)(p))->disableUpdate)
#define WINBACKGROUNDCHAR(p)       (((PW)(p))->backgroundCh)

#define WINCURX(p)                 (((PW)(p))->curX)
#define WINCURY(p)                 (((PW)(p))->curY)

#define WINXPOS(p)                 (((PW)(p))->windowarea.xLeft)
#define WINYPOS(p)                 (((PW)(p))->windowarea.yTop )
#define WINLEFT(p)                 (((PW)(p))->windowarea.xLeft)
#define WINTOP(p)                  (((PW)(p))->windowarea.yTop )
#define WINRIGHT(p)                (((PW)(p))->windowarea.xLeft + ((PW)(p))->windowarea.width  - 1)
#define WINBOTTOM(p)               (((PW)(p))->windowarea.yTop  + ((PW)(p))->windowarea.height - 1)
#define WINWIDTH(p)                (((PW)(p))->windowarea.width )
#define WINHEIGHT(p)               (((PW)(p))->windowarea.height)

#define WINRESTOREXPOS(p)          (((PW)(p))->restorearea.xLeft)
#define WINRESTOREYPOS(p)          (((PW)(p))->restorearea.yTop )
#define WINRESTORELEFT(p)          (((PW)(p))->restorearea.xLeft)
#define WINRESTORETOP(p)           (((PW)(p))->restorearea.yTop )
#define WINRESTORERIGHT(p)         (((PW)(p))->restorearea.xLeft + ((PW)(p))->restorearea.width  - 1)
#define WINRESTOREBOTTOM(p)        (((PW)(p))->restorearea.yTop  + ((PW)(p))->restorearea.height - 1)
#define WINRESTOREWIDTH(p)         (((PW)(p))->restorearea.width )
#define WINRESTOREHEIGHT(p)        (((PW)(p))->restorearea.height)

#define WINCLIENTWIDTH(p)          (((PW)(p))->clientarea.width )
#define WINCLIENTHEIGHT(p)         (((PW)(p))->clientarea.height)
#define WINCLIENTXPOS(p)           (((PW)(p))->clientarea.xLeft)
#define WINCLIENTYPOS(p)           (((PW)(p))->clientarea.yTop )
#define WINCLIENTLEFT(p)           (((PW)(p))->clientarea.xLeft)
#define WINCLIENTRIGHT(p)          (((PW)(p))->clientarea.xLeft + WINCLIENTWIDTH(p) - 1)
#define WINCLIENTTOP(p)            (((PW)(p))->clientarea.yTop )
#define WINCLIENTBOTTOM(p)         (((PW)(p))->clientarea.yTop + WINCLIENTHEIGHT(p) - 1)

#define WINFLAGS(p)                (((PW)(p))->flags)
#define WINMAGIC(p)                (((PW)(p))->magic)
#define WINID(p)                   (((PW)(p))->id)
#define WINDEFATTR(p)              (((PW)(p))->defaultAttr)
#define WINBORDERATTR(p)           (((PW)(p))->borderAttr)
#define XCLIENTTOWIN(p, x)         (WINCLIENTXPOS(p) - WINXPOS(p) + (x))
#define YCLIENTTOWIN(p, y)         (WINCLIENTYPOS(p) - WINYPOS(p) + (y))
#define WINCHARXY(p, x, y)         (WINLOGICALBUFFER(p)[((y)*WINLOGICALWIDTH(p) + (x))*2    ])
#define WINATTRXY(p, x, y)         (WINLOGICALBUFFER(p)[((y)*WINLOGICALWIDTH(p) + (x))*2 + 1])
#define WINISHIDDEN(p)             (WINFLAGS(p) & WFL_HIDDEN)
#define WINUSERDATA(p)             (((PW)(p))->userData)
#define WINCAPTION(p)              (((PW)(p))->caption)
#define WINBORDERBUFFER(p)         (((PW)(p))->borderBuffer)


#define WINVIS(p)                  (((PW)(p))->vis)
#define WINVISBUFFER(p)            (VISBUFFER(WINVIS(p)))
#define WINVISFLAGS(p)             (VISFLAGS(WINVIS(p)))
#define WINVISALIGN(p)             (VISALIGN(WINVIS(p)))
#define WINVISWIDTH(p)             (VISWIDTH(WINVIS(p)))
#define WINVISHEIGHT(p)            (VISHEIGHT(WINVIS(p)))
#define WINVISLINEFLAGS(p,y)       VISLINEFLAGS(WINVIS(p),y)
#define WINVISLINE(p, y)           ((PBYTE)&(WINVISBUFFER(p)[y*WINVISWIDTH(p)]))
#define WINVISX(p,x)               VISX(WINVIS(p),(x))
#define WINVISBYTEOFFSET(p,x)      VISBYTEOFFSET(WINVIS(p),(x))
#define WINVISBYTE(p,x,y)          VISBYTE(WINVIS(p),(x),(y))
#define WINCELLISVISIBLE(p,x,y)    (WINVISBYTE(p,x,y) & BYTEBIT(WINVISX(p,x) & 7))
#define WINWHOLELINEISVISIBLE(p,y) VISWHOLELINEISVISIBLE(WINVIS(p),(y))
#define WINWHOLELINEISHIDDEN(p,y)  VISWHOLELINEISHIDDEN(WINVIS(p),(y))

#define WINLOGICALBUFFER(p)        (((PW)p)->logicalBuffer)
#define WINLOGICALX(p)             (((PW)p)->logicalX)
#define WINLOGICALY(p)             (((PW)p)->logicalY)
#define WINLOGICALWIDTH(p)         (((PW)p)->logicalWidth)
#define WINLOGICALHEIGHT(p)        (((PW)p)->logicalHeight)

#define GetWinDirtyLoX(p) (((PW)p)->dirtyLoX)
#define GetWinDirtyHiX(p) (((PW)p)->dirtyHiX)
#define GetWinDirtyLoY(p) (((PW)p)->dirtyLoY)
#define GetWinDirtyHiY(p) (((PW)p)->dirtyHiY)

#define SetWinDirtyLoX(p,x) (((PW)p)->dirtyLoX=(x))
#define SetWinDirtyHiX(p,x) (((PW)p)->dirtyHiX=(x))
#define SetWinDirtyLoY(p,x) (((PW)p)->dirtyLoY=(x))
#define SetWinDirtyHiY(p,x) (((PW)p)->dirtyHiY=(x))


# define UNUSED_ARG(vvv) (void)(vvv)

# define MEMALLOC amc_malloc /* malloc */
# define MEMFREE  amc_free   /* free   */


/* From textwin.c */

extern HWINDOW topwindow;     // Top-most window
extern HWINDOW focuswindow;   // Window that takes keystrokes
extern HWINDOW desktopwindow; // The window that forms the background

extern int     wantBorders;
extern int     mouseKind;
extern int     defaultLogicalWidth;
extern int     defaultLogicalHeight;
extern USHORT  defaultWindowAttribute;
extern int     defaultWindowX; 
extern int     defaultWindowY; 
extern int     defaultWindowDX;
extern int     defaultWindowDY;
extern int     cascadeRangeX; 
extern int     cascadeRangeY; 
extern int     cascadeOffsetX;
extern int     cascadeOffsetY;
extern int     scrRestore;
extern int     scrMode;   
extern int     forceMono; 
extern int     forcedToMono;


extern int        screenWidth;
extern int        screenHeight;
extern int        screenCharHeight;
extern void FAR * screenBuffer;


extern int        isMonoScreen;
extern int        autoScrollToCursor;

#define MONO_ATTRIBUTE   (0x07)
#define COLOUR_ATTRIBUTE (0x17)

#define GetScreenHeight()      (screenHeight)
#define GetScreenWidth()       (screenWidth)
#define GetScreenBufferStart() (screenBuffer)


HitType WindowHitType(HWINDOW w, SHORT x, SHORT y);
void GetScreenMetrics(void);
HWINDOW WindowFromPoint(SHORT x, SHORT y);
void UpdateWindowsBelowAndIncluding(HWINDOW pWindow);

void EnsureCursorIsInWindow(HWINDOW pWindow);
void ResizeWindow(HWINDOW hWindow, SHORT dx, SHORT dy);
void SetEraseCharacter(HWINDOW w, char c);
void UpdateLogicalRect( HWINDOW hWindow,
                        int loX,
                        int hiX,
                        int loY,
                        int hiY );
void ZoomWindow( HWINDOW hWindow, HitType ht );
void ClearWindowRect(HWINDOW pWindow, SHORT x, SHORT y, SHORT w, SHORT h);


# ifdef AMC_MEMORY_PARANOIA
BOOL amc_checkwindow(PW p, int line, char * fname);
#  define IsWindow(p) ((p) && (((PW)(p))->magic == WINDOWMAGIC) && amc_checkall((PW)(p),__FILE__,__LINE__))
#  define CONSISTENCY_CHECK(p) IsWindow(p)
# else
#  define IsWindow(p) ((p) && (((PW)(p))->magic == WINDOWMAGIC))
#  define CONSISTENCY_CHECK(p)
# endif

/* From textwinc.c */

void ReadTextwinConfigFile(void);


/* From textwind.c */

void  AddLineHistory(WORD line, char * fname);
void  PrintLineHistory(void);
PVOID amc_malloc(unsigned int size);
void  amc_free(PVOID ptr);
PBYTE amc_check(PVOID ptr, int atLine, char * fname);
BOOL amc_checkall(PW p, char * file, int line);
HWINDOW FindWindowFromId(USHORT id);
HWINDOW PrimAllocWindow( USHORT x, 
                         USHORT y, 
                         USHORT w, 
                         USHORT h, 
                         int    logWidth, 
                         int    logHeight );


/* From textwini.c */

extern int bTextwinInitialized;
extern int mouseAvail;

void TextwinInit(void);
void TextwinTerm(void);


/* From textwinm.c */

void FAR mouseRoutine(void);
void MoveMouse(int mouseX, int mouseY);
void ShowMouse(void);
void HideMouse(void);
int  dotwkey(int argc, char **argv, void *p);


/* From textwint.c */

void TTYWindowWrite(HWINDOW pWindow, PSZ s);
void TTYWindowPutc(HWINDOW pWindow, char ch);



extern char far singleFrameChars[];
extern char far doubleFrameChars[];


extern int param[];


# endif /* TEXTWINP_H */

/* eof textwinp.h */

