/* VT102 terminal emulator for KA9Q - calls VT102PC.C for PC specific
 * routines
 *
 * By paul@wolf.demon.co.uk June/July 1992
 */
#include <ctype.h>

#include <conio.h>

#include "global.h"
#include "session.h"
#include "hardware.h"
#include "pc.h"
#include "vt102.h"
#include "textwin.h"
#include "emul.h"


static	void	set_flag __ARGS((struct screen * scrn, char *s,int n));
static	void	term_ident __ARGS((struct screen * scrn));
static	void	dev_status __ARGS((struct screen * scrn, char *s));
static	void	print __ARGS((struct screen * scrn, char *s));



char far vt102_british[128] = 
{
  "\x20\x01\x02\x03\x04\x05\x06\x07"
  "\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
  "\x10\x11\x12\x13\x14\x15\x16\x17"
  "\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"
  " !"
   "\""
     "$%&'()*+,-./0123456789:;<=>?"
  "@ABCDEFGHIJKLMNOPQRSTUVWXYZ["
                             "\\"
                               "]^_"
  "`abcdefghijklmnopqrstuvwxyz{|}~ "
};

char far vt102_usascii[128] = 
{
  "\x20\x01\x02\x03\x04\x05\x06\x07"
  "\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
  "\x10\x11\x12\x13\x14\x15\x16\x17"
  "\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"
  " !"
   "\""
     "#$%&'()*+,-./0123456789:;<=>?"
  "@ABCDEFGHIJKLMNOPQRSTUVWXYZ["
                             "\\"
                               "]^_"
  "`abcdefghijklmnopqrstuvwxyz{|}~ "
};

char far vt102_specialgraphics[128] = 
{
  "\x20\x01\x02\x03\x04\x05\x06\x07"
  "\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
  "\x10\x11\x12\x13\x14\x15\x16\x17"
  "\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"
  " !"
   "\""
     "#$%&'()*+,-./0123456789:;<=>?"
  "@ABCDEFGHIJKLMNOPQRSTUVWXYZ["
                             "\\"
                               "]^ "
  "heclnvٿ_ô³# "
};



void prim_vt102_putc (struct screen* scrn, char c, int echo)
{
	int num;
	int num2;
	char *cp;
	char *ap;
	register unsigned int flags;
 struct vt102_emulInfo * info = &scrn->emulInfo->vt102;


	flags = info->flags;

	if ( echo && !(flags&ECHO))
		return;

	if (info->flags & TTY)
	{
		if (c)
			TTYWindowPutc(scrn->hWnd, c);
	}


 if (c == 14)
    {
     info->GL = info->G1;
     return;
    }
 else if (c == 15)
    {
     info->GL = info->G0;
     return;
    }


	if (info->esc_ptr)
	{
		if ((c==CAN)||(c==SUB))
		{
			info->esc_ptr=0;
			return;
		}

  /* AMcC : 9Aug92 : If we've filled the escape queue, something's gone
  **                 wrong so we'll kill it.
  */
  if (info->esc_ptr >= sizeof(info->esc_queue)-2)
     {
      info->esc_ptr = 0;
      return;
     }


  cp=info->esc_queue;
		cp[info->esc_ptr++]=c;

		cp[info->esc_ptr]=0;
		if (!(flags&EMUL_ANSI))
		{
			num=1;
			switch(*(cp+1))
			{
				case 'A':Cursor_Up(scrn, 1);
						 break;
				case 'B':Cursor_Up(scrn, -1);
						 break;
				case 'C':Cursor_Right(scrn, 1);
						 break;
				case 'D':Cursor_Right(scrn, -1);
						 break;
				case 'H':set_xy(scrn, 1,1);
						 break;
				case 'I':rev_index(scrn);
						 break;
				case 'K':clr_eol(scrn);
						 break;
				case 'J':clr_eos(scrn);
						 break;
				case '<':info->flags|=EMUL_ANSI;
						 break;
				case 'W':info->flags|=PRINT_CTRL;
						 break;
				case 'X':info->flags&=(~PRINT_CTRL);
						 break;
				case 'Z':kbfeed("\x1b/Z");
						 break;
				case 'Y':if (info->esc_ptr>=4)
						 {
							set_xy(scrn, *(cp+2)-32,*(cp+3)-32);
						 }
						 else
							num=0;
						 break;
				case 0	:num=0;
						 break;
            }
			if (num)
			{
				info->esc_ptr=0;
			}
		}
		else
		{

   if (     info->esc_ptr>=2  
        &&  (c >= 0x30  &&  c <= 0x7e)
        &&  strchr("()*+",cp[1])!=NULLCHAR)
      {
       char ** ptr = NULL;

       if (cp[1] == '(')
          ptr = &info->G0;
       else
          ptr = &info->G1;

       if (ptr != NULL)
          switch (c)
                 {
                  case '0': *ptr = vt102_specialgraphics; break;
                  case 'A': *ptr = vt102_british;         break;
          
                  default:
                  case 'B': *ptr = vt102_usascii;         break;
                 }

				   info->esc_ptr=0;
      }
   else

			if ((isalpha(c))||
				((info->esc_ptr==2)&&
									(isdigit(c)||(c=='=')||(c=='>')))||
				((isdigit(c))&&(cp[1]=='#')))
			{
				if (strncmp(cp,"\x1b[",2)==0)
				{
					cp+=2;
					if (strchr("ABCDPLM",c)!=NULLCHAR)
					{
						num=atoi(cp);
						if (num==0)
							num=1;
						switch(c)
						{
							case 'A':Cursor_Up(scrn, num);
									 break;
							case 'B':Cursor_Up(scrn, -num);
									 break;
							case 'C':Cursor_Right(scrn, num);
									 break;
							case 'D':Cursor_Right(scrn, -num);
									 break;
							case 'P':del_char(scrn, num);
									 break;
							case 'L':ins_line(scrn, num);
									 break;
							case 'M':del_line(scrn, num);
									 break;
                        }
					}
					else
					{
						if (strchr("JKc",c)!=NULLCHAR)
						{
							num=atoi(cp);
							switch (c)
							{
								case 'J':switch(num)
										 {
											case 0:clr_eos(scrn);
												   break;
											case 1:clr_bos(scrn);
												   break;
											case 2:clr_scr(scrn);
												   break;
										 }
										 break;
								case 'K':switch(num)
										 {
											case 0:clr_eol(scrn);
												   break;
											case 1:clr_bol(scrn);
												   break;
											case 2:clr_line(scrn);
												   break;
										 }
										 break;
								case 'c':if ((*cp==0)||(*cp==c))
											term_ident(scrn);
										 break;
							}
						}
						else
						{
							if (strchr("Hfr",c)!=NULLCHAR)
							{
								num=atoi(cp);
								for (;((*cp!=';')&&(*cp!=c));cp++);
								cp++;
								num2=atoi(cp);
								switch(c)
								{
									case 'H':
									case 'f':set_xy(scrn, num2,num+
											((flags&COORDS_REL)?info->scroll_top:0));
											 break;

									case 'r':
            if (num <= 0)
               num = 1;

            if (num2 <= 0)
               num2 = scrn->height;

            if((num2>num)&&(num>=1)&&(num2<=scrn->height))
											 {
												info->scroll_top=num-1;
												info->scroll_bottom=num2-1;
												set_xy(scrn, 1,num);
                                             }
											 break;
								}
							}
							else
							{
								if (strchr("hlmni",c)!=NULLCHAR)
								{
									// while ((*cp)&&(*cp!=c)) : AMcC now a do loop to allow
         //                           zero-default strings
         do
									{
										ap=cp;
										for (;((*cp!=';')&&(*cp!=c));cp++);
										*(cp++)=0;
										switch(c)
										{
											case 'h':set_flag(scrn, ap,1);
													 break;
											case 'l':set_flag(scrn, ap,0);
													 break;
											case 'm':set_attr(scrn, ap);
													 break;
											case 'n':dev_status(scrn, ap);
													 break;
											case 'i':print(scrn, ap);
													 break;
										}
									} while ((*cp)&&(*cp!=c)); // AMcC : do loop, was while loop
								}
							}
						}
					}
				}
				else
				{
					if (*cp==ESC)
					{
						switch (*(cp+1))
						{
							case 'D':
          do_lf(scrn);
									 break;

							case 'E':
          console_putc(scrn, '\r');
									 do_lf(scrn);
									 break;

							case 'M':
          rev_index(scrn);
									 break;

							case '7':
          store_posn(scrn);
									 break;

							case '8':
          restore_posn(scrn);
									 break;

							case 'Z':
          term_ident(scrn);
									 break;

							case 'c':
          init_vt102(scrn);
									 break;

							case '=':
          info->flags |=  KEYPAD_APPL;
									 break;

							case '>':
          info->flags &= ~KEYPAD_APPL;
          break;

							case '#':
									 switch(*(cp+2))
									 {

										case '3':
										case '4':
										case '6':
												 set_line_attr(scrn,1);
												 break;
										case '5':
												 set_line_attr(scrn,0);
												 break;
									 }
						}
					}
				}
				info->esc_ptr=0;
			}
		}
	}
	else
	{
		if ((flags&PRINT_CTRL)&&(isprint(c)||(isspace(c))))
			print_char(info->GL[c & 0x7f]);

		switch(c)
		{
			case 0:break;
			case TAB:do_tab(scrn);
					 break;
			case ESC:info->esc_queue[0]=c;
					 info->esc_ptr=1;
					 break;
			case LF:
			case VT:
			case FF: do_lf(scrn);
					 break;
			case CAN:
			case SUB:info->esc_ptr=0;
					 break;
			default: /* do most usual case first */
					 if ( (!(flags&VT102_INSERT))||
						(!isprint(c)))
      {
       AddDirtyCurXY(scrn->hWnd);
							console_putc(scrn, info->GL[c & 0x7f]);
      }
					 else
					 {
       AddDirtyCurY(scrn->hWnd);
						ins_char(scrn);
						console_putc(scrn, info->GL[c & 0x7f]);
					 }
/*
                     if ( (!(flags&(VT102_INSERT|DH_BOTTOM|DOUBLE_WIDTH)))||
						(!isprint(c)))
						console_putc(c);
					 else
					 {
						if (flags&DH_BOTTOM)
							c=' ';
						if (flags&VT102_INSERT)
							ins_char();
						console_putc(c);
						if (flags&DOUBLE_WIDTH)
						{
							console_putc(' ');
							if (flags&VT102_INSERT)
								ins_char();
						}
					 }
*/
                     break;
		}
	}
}


void	vt102_putc (scrn,c,echo)
struct screen *scrn;
int c;  //> 2nd param changed char -> int 2.15
int echo;
{
 prim_vt102_putc(scrn,c,echo);

 UpdateDirtyArea(scrn->hWnd);
}

void	vt102_puts (scrn,s,echo)
struct screen *scrn;
char *s;
int echo;
{
	for (;*s;prim_vt102_putc(scrn,*(s++),echo));

 UpdateDirtyArea(scrn->hWnd);
}

struct	
{
 char *str;
 unsigned int mask;
} far sr_modes[] =
{
 { "4"  , VT102_INSERT },
 { "12" , ECHO         },
 { "20" , RETURN_CRLF  },
	{ "?1" , CURSOR_APPL  },
 { "?2" , EMUL_ANSI    },
 { "?6" , COORDS_REL   },
 { "?7" , WRAP         },
	{ "?18", PRINTER_FF   },
 { "?19", PRINT_FULL   },
 { "?25", CURSOR       },
   
 { NULL, 0 }
};

void	set_flag(scrn,s,n)
struct screen *scrn;
char *s;
int n;
{
	int i;
 int termWidth;

	for (i=0;sr_modes[i].mask;i++)
	{
		if (strcmp(s,sr_modes[i].str)==0)
		{
			if (n)
				scrn->emulInfo->vt102.flags|=sr_modes[i].mask;
   else
				scrn->emulInfo->vt102.flags&=(~sr_modes[i].mask);
			set_cursor(scrn, scrn->emulInfo->vt102.flags&CURSOR,scrn->emulInfo->vt102.flags&VT102_INSERT);
			return;
        }
	}

	if (strcmp(s, "?3")==0)
	{
  /* AMcC : 31Aug92 : make sure the vt102 info doesn't think
  ** that its got a larger terminal width than is available
  */

  termWidth = GetTerminalWidth(scrn->hWnd);

		if (n  &&  termWidth >= 132)
			  scrn->width = 132;
		else
			  scrn->width =  80;

  if (scrn->width > termWidth)
     scrn->width = termWidth;

		set_video_mode(scrn);
		clrscrn(scrn);
	}
}

void	dev_status(scrn, s)
struct screen *scrn;
char *s;
{
	if (strcmp(s,"6")==0)
	{
		curpos_report(scrn);
	}
}

void	term_ident(scrn)
struct screen *scrn;
{
	kbfeed("\x1b[?6c");
}

void	print(scrn, s)
struct screen *scrn;
char *s;
{
	if (strcmp(s,"4")==0)
	{
		scrn->emulInfo->vt102.flags&=(~PRINT_CTRL);
		return;
	}
	if (strcmp(s,"5")==0)
	{
		scrn->emulInfo->vt102.flags|=PRINT_CTRL;
		return;
    }
}


